#ifndef BFD_SYMBOL_H_
#define BFD_SYMBOL_H_
#include <bfd.h>
#include <map>
#include <string>
#include <cxxabi.h>

typedef std::map<bfd_vma, std::string> BFDMap;

int GetBFDSymbol(const char *fname, BFDMap& map)
{
	const int dyn = 0;
	int ret = 0;
	void *minisyms = 0;
	bfd *abfd = bfd_openr(fname, NULL);
	if (abfd == 0) return -1;
	if (bfd_check_format(abfd, bfd_object) == 0) {
		ret = -2;
		goto EXIT;
	}
	if (!(bfd_get_file_flags(abfd) & HAS_SYMS)) {
		ret = -3;
		goto EXIT;
	}
	{
		asymbol *store = bfd_make_empty_symbol(abfd);
		size_t size;
		int symnum = bfd_read_minisymbols(abfd, dyn, &minisyms, &size);
		if (symnum < 0) {
			ret = -4;
			goto EXIT;
		}
		char *p = (char*)minisyms;
		for (int i = 0; i < symnum; i++) {
			asymbol *sym = bfd_minisymbol_to_symbol(abfd, dyn, p, store);
			p += size;
			const char *name = bfd_asymbol_name(sym);
			if (name && *name == '.') continue;
			bfd_vma value = bfd_asymbol_value(sym);
			int status;
			char *demangle = abi::__cxa_demangle(name, 0, 0, &status);
			if (status == 0) {
				map[value] = std::string(demangle, strlen(demangle) - 2); /* remove "()" */
				free(demangle);
			} else {
				map[value] = name;
			}
			// printf("%s %p\n", name, value);
		}
	}
EXIT:
	free(minisyms);
	bfd_close(abfd);
	return ret;
}

static int s_callLevel = 0;
static BFDMap *s_map = 0;

const char *GetCurrentPath()
{
	static char path[4096];
	int ret = readlink("/proc/self/exe", path, sizeof(path));
	return (ret != -1) ? path : 0;
}

int InitBFDSymbol()
__attribute__((constructor));

int InitBFDSymbol()
{
	static BFDMap map;
	const char *path = GetCurrentPath();
	if (path == 0) return -1;
	int ret = GetBFDSymbol(path, map);
	if (ret) return ret;
	s_map = &map;
	return 0;
}

const char *GetSymbolString(void *addr)
{
	return s_map ? (*s_map)[(bfd_vma)addr].c_str() : 0;
}

extern "C" {

void __cyg_profile_func_enter(void *funcAddr, void *callSite)
{
//	printf("enter=%p, from %p\n", funcAddr, callSite);
	if (s_map) {
		for (int i = 0; i < s_callLevel; i++) {
			printf("  ");
		}
		printf("in %s [%p]\n", (*s_map)[(bfd_vma)funcAddr].c_str(), funcAddr);
		s_callLevel++;
	}
}

void __cyg_profile_func_exit(void *funcAddr, void *callSite)
{
//	printf("leave=%p, from %p\n", funcAddr, callSite);
	s_callLevel--;
}

}

#endif /* BFD_SYMBOL_H_ */

